
public static final int OFFSET_H = 15;
public static final int OFFSET_W = 30;
public static final int LEGEND_H = 80;
public static final int METABOLITE_H = 1;
public static final int GENE_H = 5;
public static final int GENE_W = 6;
public static final int SPACE_BETWEEN_NODES = 2;
public static final int SPACE_BETWEEN_METABOLITES = 1;
public static final int BREAK_SPACE = 16;

public final color PATHWAY_COLOR = color( 235 );
public final color PATHWAY_LINE_COLOR = color( 215 );
public final color NO_CORRELATION_COLOR = color( 200 );
public final color GENE_COLOR = color( 0 );
public static final int GENE_FILL_ALPHA = 50;
public final color METABOLITE_COLOR = color( 150 );
public final color SELECTED_NODE_COLOR = color( 255, 50, 30 );
public final color BRANCH_GRAY_COLOR = color( 150 );
public final color SCROLL_LINE_COLOR = color( 215 );
public final color LABEL_COLOR = #ff7f00;
public final color LABEL_COLOR2 = #377db8;

public static final int BRANCH_LINE_WIDTH = 2;
public static final int BRANCH_LINE_BRANCH_WIDTH = 1;
public static final int BRANCH_LEVEL_OFFSET = 2;
public static final int BRANCH_SPACE = 4;
public static final int BRANCH_BRANCH_WIDTH = 14;
public static final int CYCLE_ICON_WIDTH = 8;

public static final int TEXT_HEIGHT = 8;
public static final int AVE_TEXT_WIDTH = 6;

public static final int LEGEND_ELEMENT_OFFSET = 15;
public static final int SCROLLBAR_W = 10;
public static final int EXTRA_FOR_SCROLLING = 4;

int[] _selected_node;
int _num_nodes_selected;
int _name_center_x, _pathway_center_x;
boolean[][] _paralogs;
boolean _correlation_rolled_over;
int _rolled_over_correlation;

void setParalogFlags()
{
  _paralogs = new boolean[_species.length][_num_nodes_selected+1];

  // initialize to false
  for ( int i = 0; i < _species.length; i++ )
    for ( int j = 0; j < _num_nodes_selected+1; j++ )
      _paralogs[i][j] = false;

  if ( _num_nodes_selected == 1 ) return;

  // set the flags
  for ( int i = 0; i < _species.length; i++ )
    for ( int j = 0; j < _num_nodes_selected-1; j++ )
    {
      for ( int k = j+1; k < _num_nodes_selected; k++ )
      {
        if ( !_pathway.getNode(_selected_node[j]).paralogName(i).equalsIgnoreCase("NONE") && !_paralogs[i][k] )
        {
          if ( _pathway.getNode(_selected_node[j]).paralogName(i).equals(_pathway.getNode(_selected_node[k]).paralogName(i)) )
            _paralogs[i][j] = _paralogs[i][k] = true;
        }
      }
    }
}

class CorrelationView
{
  int _width, _height, _pathway_width, _pathway_height;
  int _pathway_origin_x, _pathway_origin_y, _legend_origin_x, _legend_origin_y;
  int _pathway_start_y, _pathway_stop_y, _pathway_screen_height;
  int _scrollbar_x, _scrollbar_length, _scroll_length, _scroll_offset;
  int _pathway_offset;
  int _legend_height;
  boolean _use_scroll, _scrollbar_pressed;
  int _scrollbar_pressed_y, _extra_bit;

  boolean _node_rolled_over;
  int _rolled_over_node;
  int _rolled_over_class;

  int _num_corr_selected;
  int _single_selected_corr;

  CorrelationView()
  {
    // interaction initialization
    _node_rolled_over = false;
    _selected_node = new int[UPPER_BOUND_MAX_NUM_SELECTION];
    _selected_node[0] = 0;
    _num_nodes_selected = 1;

    _num_corr_selected = 1;
    _single_selected_corr = 0;
    
    _correlation_rolled_over = false;
    _rolled_over_correlation = 0;

    setParalogFlags();
  }



  void numCorrToRender( int num )
  {
    _num_corr_selected = min(num, _pathway.numCorrelation());
  }

  void singleSelectedCorr( int index )
  {
    if ( index < _pathway.numCorrelation() ) _single_selected_corr = index;
  }

  void setDimensions( int w, int h )
  { 
    _width = w;
    _height = h;

    // legend height
    _legend_height =  LEGEND_ELEMENT_OFFSET*(3+_pathway.numCorrelation());

    _pathway_width = _width - 4*OFFSET_W;
    // for now, compute height based on numbe of nodes
    // NOTE: will need to compute the max height and add scrolling!!!!
    //_pathway_height = _height - 2*OFFSET_H - LEGEND_H; // max height
    _pathway_height = _pathway.numGenes()*(GENE_H+SPACE_BETWEEN_NODES);
    _pathway_height += _pathway.numMetabolites()*(METABOLITE_H+SPACE_BETWEEN_NODES);
    _pathway_height += _pathway.numMetabolites()*(METABOLITE_H+SPACE_BETWEEN_METABOLITES)*(_pathway.numCorrelation()-1);
    _pathway_height += _pathway.numBreakPoints()*(BREAK_SPACE-SPACE_BETWEEN_NODES);
    _pathway_height += _pathway.numBranchPoints()*(BRANCH_SPACE-SPACE_BETWEEN_NODES);
    _pathway_height -= SPACE_BETWEEN_NODES;
    _pathway_height -= (METABOLITE_H%2);

    _pathway_origin_x = 3*OFFSET_W;
    _pathway_origin_y = OFFSET_H;
    _pathway_center_x = _pathway_origin_x + _pathway_width/2;
    _name_center_x = OFFSET_W;

    _legend_origin_x = 0;
    _legend_origin_y = min((_pathway_origin_y + 2*OFFSET_H + _pathway_height), (_height-_legend_height));

    // set the positions for the nodes
    setNodePositions();

    // set the positions for the pathway names
    setPathwayNamePositions();

    // set the branch positions
    setBranchPositions();

    // set the cycle positions
    setCyclePositions();

    _scrollbar_pressed = false;
    _pathway_screen_height = _height - _legend_height - 3*OFFSET_H;
    if ( _pathway_screen_height < _pathway_height ) _use_scroll = true;
    else
    {
      _use_scroll = false;
      _pathway_screen_height = _pathway_height;
    }

    _extra_bit = _pathway.getPathway(_pathway.numPathways()-1).renderY();
    _extra_bit += AVE_TEXT_WIDTH*_pathway.getPathway(_pathway.numPathways()-1).name().length();
    _extra_bit -= (_pathway_height + _pathway_origin_y);
    _extra_bit = max( EXTRA_FOR_SCROLLING, _extra_bit );



    _pathway_start_y = _pathway_origin_y - EXTRA_FOR_SCROLLING;
    _pathway_stop_y = _pathway_origin_y + _pathway_screen_height + EXTRA_FOR_SCROLLING;
    _scrollbar_x = _width;
    _scrollbar_length = _pathway_stop_y-_pathway_start_y;
    _scroll_length = (int)((float)_scrollbar_length*(float)_pathway_screen_height/(float)(_pathway_height+EXTRA_FOR_SCROLLING+_extra_bit));
    _scroll_offset = 0;

    computePathwayOffset();
  }

  void computePathwayOffset()
  {
    if ( _use_scroll )
    {
      _pathway_offset = (int)((float)(_scroll_offset)/(float)(_scrollbar_length-_scroll_length) * (float)(_pathway_height+EXTRA_FOR_SCROLLING+_extra_bit - _pathway_screen_height));
    }
    else _pathway_offset = 0;
  }

  void setNodePositions()
  {
    PathwayNode node;
    float corr_min = _pathway.correlationMin();
    float corr_max = _pathway.correlationMax();
    float corr_diff = corr_max - corr_min;  
    int y = _pathway_origin_y;
    int x = _pathway_origin_x; 
    for ( int i = 0; i < _pathway.numNodes(); i++ )
    {
      node = _pathway.getNode(i);
      for ( int c = 0; c < _pathway.numCorrelation(); c++ )
      {
        // check if we have a correlation value
        if ( node.haveCorrelation(c) ) 
        {
          // compute the x location based on the correlation value
          x = _pathway_origin_x + round(((node.correlation(c)-corr_min)/corr_diff) * _pathway_width);  

          node.setXY( x, y, c ); 
        }
        else
        { 
          // set off to the left
          if ( node.type() == GENE )
            x = _pathway_origin_x - GENE_W/2;
          else if ( node.type() == METABOLITE )
            x = _pathway_origin_x - GENE_W - 1;

          node.setXY( x, y, c );
        }

        // if this is a metabolite AND there are multiple metrics, move the y for the next metric
        if ( (node.type() == METABOLITE) && (c < (_pathway.numCorrelation()-1)) ) y += METABOLITE_H+SPACE_BETWEEN_METABOLITES;
      }

      // move the y coordinate down....
      if ( node.type() == GENE ) y += GENE_H+SPACE_BETWEEN_NODES;
      else y += METABOLITE_H+SPACE_BETWEEN_NODES;

      // check for the break points
      if ( node.branchPoint() )
        y += (BRANCH_SPACE-SPACE_BETWEEN_NODES);
      if ( node.breakPoint() ) 
        y += (BREAK_SPACE-SPACE_BETWEEN_NODES);
    }
  }

  void setPathwayNamePositions()
  {
    // set the rendering locations for the pathway names
    for ( int i = 0; i < _pathway.numPathways(); i ++ )
    {
      // for the pathway names
      _pathway.getPathway(i)._render_y = _pathway.getNode(_pathway.getPathway(i).start()).y(0);
      _pathway.getPathway(i)._render_x = _name_center_x - TEXT_HEIGHT/2 - 2;
    }
  }

  void setBranchPositions()
  {
    // set the rendering locations for the pathway names
    int start_y1, start_y2, stop_y;
    for ( int i = 0; i < _pathway.numBranches(); i ++ )
    {
      start_y1 = _pathway.getNode(_pathway.getBranch(i).start1()).y(0);
      start_y2 = _pathway.getNode(_pathway.getBranch(i).start2()).y(0);

      stop_y = _pathway.getNode(_pathway.getBranch(i).stop()).y(0);
      if( _pathway.getNode(_pathway.getBranch(i).stop()).type() == GENE ) stop_y += GENE_H;
      else stop_y += METABOLITE_H;

      _pathway.getBranch(i)._render_y_start1 = start_y1;
      _pathway.getBranch(i)._render_y_end1 = start_y2 - BRANCH_SPACE;
      _pathway.getBranch(i)._render_y_start2 = start_y2;     
      _pathway.getBranch(i)._render_y_end2 = stop_y;  

      // for the branch names
      _pathway.getBranch(i)._render_y1 = start_y1;
      _pathway.getBranch(i)._render_y2 = start_y2;
      _pathway.getBranch(i)._render_x1 = _name_center_x + TEXT_HEIGHT/2 + 2;
      _pathway.getBranch(i)._render_x2 = _name_center_x + TEXT_HEIGHT/2 + 2;
    }
  }

  void setCyclePositions()
  {
    // set the rendering locations for the pathway names
    int start_y, stop_y;
    for ( int i = 0; i < _pathway.numCycles(); i ++ )
    {
      start_y = _pathway.getNode(_pathway.getCycle(i).start()).y(0);
      stop_y = _pathway.getNode(_pathway.getCycle(i).stop()).y(0);
      if( _pathway.getNode(_pathway.getCycle(i).stop()).type() == GENE ) stop_y += GENE_H;
      else stop_y += METABOLITE_H;

      _pathway.getCycle(i)._render_y_start = start_y;
      _pathway.getCycle(i)._render_y_end = stop_y;  

      // for the cycle names
      _pathway.getCycle(i)._render_y = start_y;
      _pathway.getCycle(i)._render_x = _name_center_x + TEXT_HEIGHT/2 + 2;
    }
  }

  void render()
  {    
    noSmooth();

    //
    // NOTE: the screen coordinate system starts in the upper right corner,
    //   so we need to render with this in mind!!!
    //

    strokeCap(SQUARE);

    //-----------------------------------------------------------------------
    // render the underlying pathway box with outline and line up middle   
    noStroke();    
    fill( PATHWAY_COLOR );
    rectMode( CORNER );
    rect( _pathway_origin_x, _pathway_origin_y-_pathway_offset, _pathway_width, _pathway_height+1 );

    float offset = floor( METABOLITE_H/2 );
    stroke( PATHWAY_LINE_COLOR );
    strokeWeight( METABOLITE_H );
    line( _pathway_origin_x+offset, _pathway_origin_y-_pathway_offset, _pathway_origin_x+offset, _pathway_origin_y+_pathway_height-_pathway_offset );
    line( _pathway_origin_x+_pathway_width/2, _pathway_origin_y-_pathway_offset, _pathway_origin_x+_pathway_width/2, _pathway_origin_y+_pathway_height-_pathway_offset );
    line( _pathway_origin_x+_pathway_width-offset, _pathway_origin_y-_pathway_offset, _pathway_origin_x+_pathway_width-offset, _pathway_origin_y+_pathway_height-_pathway_offset );

    if ( _use_scroll )
    {
      //-----------------------------------------------------------------------
      // render the scroll bar
      stroke( SCROLL_LINE_COLOR );
      strokeWeight( 1 );
      fill( PATHWAY_COLOR );
      rectMode( CORNER );
      rect( _scrollbar_x, _pathway_start_y+_scroll_offset, SCROLLBAR_W, _scroll_length );

      line( _scrollbar_x, _pathway_start_y, _scrollbar_x, _pathway_stop_y );
      line( _scrollbar_x+SCROLLBAR_W, _pathway_start_y, _scrollbar_x+SCROLLBAR_W, _pathway_stop_y );
    }

    //-----------------------------------------------------------------------
    // render the pathway
    PathwayNode node;
    color node_color;
    boolean enlarge;
    int c;
    int selected_name_bottom_coord = 0-_pathway_offset;
    for ( int j = 0; j < _num_corr_selected; j++ )
    {
      if ( _num_corr_selected == 1 ) c = _single_selected_corr;
      else c = j;
      for ( int i = 0; i < _pathway.numNodes(); i++ )
      {
        node = _pathway.getNode(i);

        // render the underlying metabolite gray
        if ( node.type() == METABOLITE )
        {
          offset = floor( METABOLITE_H/2 );

          if ( (c == 0) || (_num_corr_selected == 1) )
          {
            strokeWeight( METABOLITE_H );
            stroke( PATHWAY_LINE_COLOR );
            line( _pathway_origin_x, node.y(0)+offset-_pathway_offset, _pathway_origin_x+_pathway_width, node.y(0)+offset-_pathway_offset );
          }
        }

        // check if we have a correlation value
        if ( node.haveCorrelation(c) ) 
        {
          // render the node
          if ( node.type() == GENE )
          {
            if ( _node_rolled_over && (_rolled_over_node == i) && (_rolled_over_class == c) ) enlarge = true;
            else enlarge = false;

            int bar_length = 0;
            if ( (c == 0) && (_num_corr_selected > 1) && node.haveCorrelation(1) ) bar_length = node.x(1)-node.x(0)+GENE_H/2;

            renderGene( node.x(c), node.y(c)-_pathway_offset, node.direction(), isSelected(i), enlarge, c, bar_length );
          }
          else
          {
            if ( _node_rolled_over && (_rolled_over_node == i) && (_rolled_over_class == c) ) enlarge = true;
            else enlarge = false;

            renderMetabolite( node.x(c), node.y(c)-_pathway_offset, isSelected(i), enlarge, c );
          }  

          // if the node is selected, render the label
          if ( isSelected(i) && (c==0) )
          {
            boolean is_rolled_over = false;
            for ( int k = 0; k < _num_nodes_selected; k++ )
              if ( (_selected_node[k] == i) && _selected_node_rolled_over[k] ) is_rolled_over = true;

            // render the selection flags
            rectMode( CORNER );
            noStroke();
            fill( SELECTED_NODE_COLOR );
            if ( node.type() == GENE )
              rect( _pathway_origin_x+_pathway_width+2, node.y(0)+1-_pathway_offset, 4, 2 );
            else
              rect( _pathway_origin_x+_pathway_width+2, node.y(0)-_pathway_offset, 4, 2 );

            textAlign( LEFT, CENTER );
            if ( is_rolled_over ) 
            {
              textFont(_plotFont_8b); 
              fill( PHYLO_TREE_COLOR_ROLLED_OVER );
            }
            else
            {
              textFont(_plotFont_8); 
              fill( TEXT_COLOR );
            }
            text( node.shortName(), _pathway_origin_x+_pathway_width+8, max(node.y(0)-_pathway_offset, selected_name_bottom_coord) );
            selected_name_bottom_coord = max(node.y(0)-_pathway_offset, selected_name_bottom_coord) + TEXT_HEIGHT+1;
          }  
        }
        /*
        else
         { 
         // render the node
         if ( node.type() == GENE )
         renderGene( node.x(c), node.y(c), node.direction(), false, false, 0, 0 );
         else if ( node.type() == METABOLITE )
         renderMetabolite( node.x(c), node.y(c), false, false, c );
         }*/

        // render the break points
        if ( node.branchPoint() ) 
        {
          noStroke();
          fill( BACKGROUND_COLOR );
          rectMode( CORNER );
          if ( node.type() == GENE ) rect( _pathway_origin_x-METABOLITE_H, node.y(c)+GENE_H-_pathway_offset, _pathway_width+2*METABOLITE_H, BRANCH_SPACE );
          else rect( _pathway_origin_x-METABOLITE_H, node.y(_pathway.numCorrelation()-1)+METABOLITE_H-_pathway_offset, _pathway_width+2*METABOLITE_H, BRANCH_SPACE );
        }
        if ( node.breakPoint() ) 
        {
          noStroke();
          fill( BACKGROUND_COLOR );
          rectMode( CORNER );
          if ( node.type() == GENE ) rect( _pathway_origin_x-METABOLITE_H, node.y(c)+GENE_H-_pathway_offset, _pathway_width+2*METABOLITE_H, BREAK_SPACE );
          else rect( _pathway_origin_x-METABOLITE_H, node.y(_pathway.numCorrelation()-1)+METABOLITE_H-_pathway_offset, _pathway_width+2*METABOLITE_H, BREAK_SPACE );
        }
      }
    }

    //-----------------------------------------------------------------------
    // write the pathway landmark names and pathway name
    textFont(_plotFont_8); 
    int x, y, stop_y = -1-_pathway_offset;
    boolean n_offset = false;
    for ( int i = 0; i < _pathway.numPathways(); i++ )
    {
      textAlign( RIGHT, CENTER );
      fill( TEXT_COLOR );

      node = _pathway.getNode(_pathway.getPathway(i).start());
      if ( !node.branchPoint() && !node.cyclePoint() )    
        text( node.shortName(), _pathway_origin_x - 2, node.y(0)-_pathway_offset );
      node = _pathway.getNode(_pathway.getPathway(i).stop());
      if ( !node.branchPoint() && !node.cyclePoint() )   
        text( node.shortName(), _pathway_origin_x - 2, node.y(0)-_pathway_offset );

      if ( !_pathway.getPathway(i).haveName() ) continue;

      x = _pathway.getPathway(i).renderX();
      y = _pathway.getPathway(i).renderY()-_pathway_offset;
      if ( (stop_y) >= y && !n_offset )
      { 
        x -= (TEXT_HEIGHT + 3);
        n_offset = true;
      }
      else n_offset = false;

      pushMatrix();
      translate( x, y );
      rotate( -PI/2 );

      fill( TEXT_COLOR );
      textAlign( RIGHT, BOTTOM );
      text( _pathway.getPathway(i).name().toUpperCase(), 0, 0 );

      popMatrix();

      // get the start and stop of the line
      stop_y = max(_pathway.getNode(_pathway.getPathway(i).stop()).y(0)-_pathway_offset, y+AVE_TEXT_WIDTH*_pathway.getPathway(i).name().length());

      // there is a line to render
      strokeWeight(1);
      stroke( BRANCH_GRAY_COLOR );
      line( x, y, x, stop_y );
    }

    //-----------------------------------------------------------------------
    // render the branches
    Branch b;
    PathwayNode n;
    int nest_offset;
    textFont(_plotFont_8);     
    stop_y = -1-_pathway_offset;
    n_offset = false;
    for ( int i = 0; i < _pathway.numBranches(); i ++ )
    {
      b = _pathway.getBranch(i);

      // render the lines down the sides
      nest_offset = b.nestLevel()*(BRANCH_LEVEL_OFFSET+BRANCH_LINE_WIDTH);
      x = _pathway_origin_x - METABOLITE_H - BRANCH_LEVEL_OFFSET - nest_offset;  

      stroke( BRANCH_GRAY_COLOR );
      strokeWeight( BRANCH_LINE_WIDTH );
      line( x, b.renderYStart1()-_pathway_offset, x, b.renderYEnd1()-_pathway_offset );
      line( x, b.renderYStart2()-_pathway_offset, x, b.renderYEnd2()-_pathway_offset ); 

      // render the branch branches
      if ( (b.start1() != 0) && !_pathway.getNode(b.start1()-1).breakPoint() ) 
      {
        n = _pathway.getNode(b.start1()-1);
        y = n.y(0)-_pathway_offset;
        if ( n.type() == GENE ) y += GENE_H;
        else y += METABOLITE_H;

        y -= BRANCH_LINE_BRANCH_WIDTH;
      }
      else y = _pathway.getNode(b.start1()).y(0)-_pathway_offset;

      strokeWeight( BRANCH_LINE_BRANCH_WIDTH );
      line( x, y, x-BRANCH_BRANCH_WIDTH, y );
      line( x-BRANCH_BRANCH_WIDTH/2, y, x-BRANCH_BRANCH_WIDTH/2, b.renderYStart1()-_pathway_offset );
      line( x-BRANCH_BRANCH_WIDTH/2, b.renderYStart1()-_pathway_offset, x-BRANCH_BRANCH_WIDTH/4, b.renderYStart1()-_pathway_offset );
      line( x-BRANCH_BRANCH_WIDTH, y, x-BRANCH_BRANCH_WIDTH, b.renderYStart2()-_pathway_offset );
      line( x-BRANCH_BRANCH_WIDTH, b.renderYStart2()-_pathway_offset, x-BRANCH_BRANCH_WIDTH/4, b.renderYStart2()-_pathway_offset );

      // render the branch letters
      textAlign( RIGHT, TOP );
      fill( BRANCH_GRAY_COLOR );
      text( b.branch1Letter(), x-BRANCH_BRANCH_WIDTH/4, b.renderYStart1()-_pathway_offset );
      text( b.branch2Letter(), x-BRANCH_BRANCH_WIDTH/4, b.renderYStart2()-_pathway_offset );

      // render the branch landmark names
      textAlign( RIGHT, CENTER );
      fill( TEXT_COLOR );

      int index = b.start1()-1;
      for ( int j = 0; j < _pathway.numPathways(); j++ ) 
        if ( index == _pathway.getPathway(j).start() ) index = -1;
      if ( index != -1) 
        text( _pathway.getNode(index).shortName(), x- BRANCH_BRANCH_WIDTH - 1, _pathway.getNode(index).y(0)-_pathway_offset );

      index = b.stop()+1;
      for ( int j = 0; j < _pathway.numPathways(); j++ ) 
        if ( index == _pathway.getPathway(j).stop() ) index = _pathway.numNodes();
      if ( index != _pathway.numNodes() )
        text( _pathway.getNode(index).shortName(), _pathway_origin_x - 2 - nest_offset, _pathway.getNode(index).y(0)-_pathway_offset );

      // render the branch names
      fill( TEXT_COLOR );
      textAlign( RIGHT, BOTTOM );
      if ( _pathway.getBranch(i).haveName1() )
      {
        x = _pathway.getBranch(i).renderX1();
        y = _pathway.getBranch(i).renderY1()-_pathway_offset;
        if ( (stop_y) >= y && !n_offset )
        { 
          x += (TEXT_HEIGHT + 3);
          n_offset = true;
        }
        else n_offset = false;

        pushMatrix();
        translate( x, y );
        rotate( -PI/2 );

        text( _pathway.getBranch(i).branch1Name()/*.toUpperCase()*/, 0, 0 );

        popMatrix();

        // get the start and stop of the line
        stop_y = max(_pathway.getNode(_pathway.getBranch(i).stop()).y(0)-_pathway_offset, y+AVE_TEXT_WIDTH*_pathway.getBranch(i).branch1Name().length());

        // there is a line to render
        strokeWeight(1);
        stroke( BRANCH_GRAY_COLOR );
        line( x, y, x, stop_y );
      }

      if ( _pathway.getBranch(i).haveName2() )
      {
        x = _pathway.getBranch(i).renderX2();
        y = _pathway.getBranch(i).renderY2()-_pathway_offset;
        if ( (stop_y) >= y && !n_offset )
        { 
          x += (TEXT_HEIGHT + 3);
          n_offset = true;
        }
        else n_offset = false;

        pushMatrix();
        translate( x, y );
        rotate( -PI/2 );

        text( _pathway.getBranch(i).branch2Name()/*.toUpperCase()*/, 0, 0 );

        popMatrix();

        // get the start and stop of the line
        stop_y = max(_pathway.getNode(_pathway.getBranch(i).stop()).y(0)-_pathway_offset, y+AVE_TEXT_WIDTH*_pathway.getBranch(i).branch2Name().length());

        // there is a line to render
        strokeWeight(1);
        stroke( BRANCH_GRAY_COLOR );
        line( x, y, x, stop_y );
      }
    }

    //-----------------------------------------------------------------------
    // render the cycles
    Cycle cyc;
    textFont(_plotFont_8);     
    stop_y = -1-_pathway_offset;
    n_offset = false;
    for ( int i = 0; i < _pathway.numCycles(); i ++ )
    {
      cyc = _pathway.getCycle(i);

      // render the lines down the sides
      x = _pathway_origin_x - METABOLITE_H - BRANCH_LEVEL_OFFSET;  

      stroke( BRANCH_GRAY_COLOR );
      strokeWeight( BRANCH_LINE_WIDTH );
      line( x, cyc.renderYStart()-_pathway_offset+CYCLE_ICON_WIDTH/2, x, cyc.renderYEnd()-_pathway_offset );

      // render the cycles icon
      fill( BACKGROUND_COLOR );
      //strokeWeight( 1 );
      arc( x-CYCLE_ICON_WIDTH/2, cyc.renderYStart()-_pathway_offset+CYCLE_ICON_WIDTH/2, CYCLE_ICON_WIDTH, CYCLE_ICON_WIDTH, PI/2, TWO_PI );


      // render the branch landmark names
      textAlign( RIGHT, CENTER );
      fill( TEXT_COLOR );
      text( _pathway.getNode(cyc.start()).shortName(), x-  CYCLE_ICON_WIDTH -1, _pathway.getNode(cyc.start()).y(0)-_pathway_offset );
      text( _pathway.getNode(cyc.stop()).shortName(), x - 2, _pathway.getNode(cyc.stop()).y(0)-_pathway_offset );

      // render the cycle name
      fill( TEXT_COLOR );
      textAlign( RIGHT, BOTTOM );
      if ( _pathway.getCycle(i).haveName() )
      {
        x = _pathway.getCycle(i).renderX();
        y = _pathway.getCycle(i).renderY()-_pathway_offset;
        if ( (stop_y) >= y && !n_offset )
        { 
          x += (TEXT_HEIGHT + 3);
          n_offset = true;
        }
        else n_offset = false;

        pushMatrix();
        translate( x, y );
        rotate( -PI/2 );

        text( _pathway.getCycle(i).cycleName(), 0, 0 );

        popMatrix();

        // get the start and stop of the line
        stop_y = max(_pathway.getNode(_pathway.getCycle(i).stop()).y(0)-_pathway_offset, y+AVE_TEXT_WIDTH*_pathway.getCycle(i).cycleName().length());

        // there is a line to render
        strokeWeight(1);
        stroke( BRANCH_GRAY_COLOR );
        line( x, y, x, stop_y );
      }
    }

    if ( _use_scroll )
    {
      //-----------------------------------------------------------------------
      // cover up the extra pathway
      fill( BACKGROUND_COLOR );
      noStroke();
      rectMode( CORNERS );
      rect( 0, -3*BORDER_WIDTH, _w, _pathway_start_y );
      rect( 0, _pathway_stop_y, _w, _height+BORDER_WIDTH );

      //-----------------------------------------------------------------------
      // render the scroll bar

      stroke( SCROLL_LINE_COLOR );
      strokeWeight( 1 );
      line( 0, _pathway_start_y, _scrollbar_x+SCROLLBAR_W, _pathway_start_y );
      line( 0, _pathway_stop_y, _scrollbar_x+SCROLLBAR_W, _pathway_stop_y );      
    }

    //-----------------------------------------------------------------------
    // render the axis labels  
    fill( TEXT_COLOR );
    textFont(_plotFont_8); 
    stroke( SCROLL_LINE_COLOR );
    strokeWeight( 1 );
    int l_offset = 5;

    String s1, s3;
    float corr_min = _pathway.correlationMin();
    float corr_max = _pathway.correlationMax();
    float corr_diff = corr_max - corr_min;
    int digis=0;
    if ( abs(corr_min) < 10 ) digis = 1;
    else if ( abs(corr_min) < 100 ) digis = 2;
    else if ( abs(corr_min) < 1000 ) digis = 3;
    else if ( abs(corr_min) < 10000 ) digis = 4;
    s1 = nf( corr_min, digis, 2 );
    if ( abs(corr_max) < 10 ) digis = 1;
    else if ( abs(corr_max) < 100 ) digis = 2;
    else if ( abs(corr_max) < 1000 ) digis = 3;
    else if ( abs(corr_max) < 10000 ) digis = 4;
    s3 = nf( corr_max, digis, 2 );

    x = _pathway_origin_x;
    y = _pathway_start_y-l_offset;
    textAlign( CENTER, BOTTOM );
    text( s1, x, y );
    text( s3, x + _pathway_width, y );
    line( x, y, x, y+l_offset );
    line( x + _pathway_width, y, x + _pathway_width, y+l_offset );

    y = _pathway_stop_y+l_offset;
    textAlign( CENTER, TOP );
    text( s1, x, y );
    text( s3, x + _pathway_width, y );
    line( x, y, x, y-l_offset );
    line( x + _pathway_width, y, x + _pathway_width, y-l_offset );

    //-----------------------------------------------------------------------
    // render the symbol legend
    renderLegend();



    //-----------------------------------------------------------------------
    // if the mouse is hoovering over a node, then render something  
    if ( _node_rolled_over )
    {
      node = _pathway.getNode(_rolled_over_node);

      s1 = nf( node.correlation(_rolled_over_class), digis, 2 );
      int box_h = TEXT_HEIGHT+2;
      int box_w;
      if ( node.name().equals(node.shortName()) )   
        box_w = AVE_TEXT_WIDTH*(node.name().length()+s1.length()+1);
      else
        box_w = AVE_TEXT_WIDTH*(node.name().length()+s1.length()+1+node.shortName().length()+3);
      int mx = mouseX - _correlation_view_origin[0] + 5;
      int my = mouseY - _correlation_view_origin[1] - 5;

      noStroke();
      //fill( PATHWAY_COLOR, 200 );
      fill( color(255), 150 );
      rectMode( CORNER );
      rect( mx, my, box_w, -box_h );

      textAlign( LEFT, BOTTOM );
      fill( color(0) );   
      if ( node.name().equals(node.shortName()) )   
        text(node.name() + "  " + s1,  mx+2, my+1 );
      else
        text(node.shortName() + " / " + node.name() + "  " + s1,  mx+2, my+1 );
    }  


  }

  void renderMetabolite( int x, int y, boolean selected, boolean enlarge, int corr_order_num )
  {   
    int offset = floor( METABOLITE_H/2 );

    stroke( METABOLITE_COLOR );
    if (enlarge) strokeWeight( 2*METABOLITE_H );
    else strokeWeight( METABOLITE_H );
    line( _pathway_origin_x, y+offset, x, y+offset );
  }

  void renderGene( int x, int y, int direction, boolean selected, boolean enlarge, int corr_order_num, int bar_length )
  {  
    color c = METABOLITE_COLOR;
    switch ( direction )
    {
    case FORWARD:   
      c = SELECTION_COLOR[0]; 
      break;
    case REVERSE:
      c = SELECTION_COLOR[1]; 
      break;
    case BIDIRECTIONAL:
      c = SELECTION_COLOR[2]; 
      break;   
    case LEGEND:
      c = METABOLITE_COLOR;
      break;
    }

    // render the bar
    boolean fill_with_color = true;
    if ( bar_length != 0 )
    {
      noStroke();
      fill( c, GENE_FILL_ALPHA );
      rectMode( CORNER );
      rect( x, y, bar_length-1, GENE_H );

      fill_with_color = false;
    }

    if (enlarge) strokeWeight( 2 );
    else strokeWeight( 1 );
    stroke( c );

    switch ( corr_order_num )
    {
    case 0:
      fill( c, GENE_FILL_ALPHA );
      if ( enlarge )
        ellipse( x, y+GENE_H/2, GENE_H+1, GENE_H+1 );
      else
        ellipse( x, y+GENE_H/2, GENE_H-1, GENE_H-1 );
      break;
    case 1:
      noFill();
      if (enlarge) 
      {
        line( x-GENE_H, y+GENE_H/2, x+GENE_H, y+GENE_H/2 );
        line( x, y-GENE_H/2, x, y+GENE_H*3/2-1 );
      }
      else
      {
        line( x-GENE_H/2, y+GENE_H/2, x+GENE_H/2, y+GENE_H/2 );
        line( x, y, x, y+GENE_H-1 );
      }
      break;
    case 2:
      if (enlarge) strokeWeight( 4 );
      else strokeWeight( 2 );
      if (enlarge)
        line( x, y-2, x, y+GENE_H+1 );
      else
        line( x, y, x, y+GENE_H-1 );
      break;
    case 3:
      rectMode( CENTER );
      fill(c);
      rect(  x, y+GENE_H/2, GENE_H-1, GENE_H-1 );
      break;
    }
  }

  void renderLegend()
  {

    /*rectMode( CORNER );
     noFill();
     stroke( color(0) );
     rect( _legend_origin_x, _legend_origin_y, _width, _legend_height );*/

    int x = _legend_origin_x; 
    int y = _legend_origin_y-LEGEND_ELEMENT_OFFSET/4;

    textFont(_veranda_small);
    textAlign( LEFT, TOP );
    fill( LABEL_COLOR );
    text( "KEY", x-10, y );
    stroke( LABEL_COLOR );
    strokeWeight( 1 );
    line( x-10, y-2, x+_width+10, y-2 );

    x += 20;
    //y += LEGEND_ELEMENT_OFFSET/2;

    textFont(_plotFont_8b);
    textAlign( LEFT, TOP );
    fill( TEXT_COLOR );
    text( "Genes", x, y );


    x += GENE_W/2; 
    y +=LEGEND_ELEMENT_OFFSET+4;

    textFont(_plotFont_8); 

    // render the symbol legend
    textAlign( LEFT, CENTER );

    renderGene( x, y, FORWARD, false, false, 3, 0 );
    fill( TEXT_COLOR );
    text( "forward", x + GENE_W/2 + 3, y );

    x += 57;
    renderGene( x, y, REVERSE, false, false, 3, 0 );
    fill( TEXT_COLOR );
    text( "reverse", x + GENE_W/2 + 3, y );

    x +=  57;
    renderGene( x, y, BIDIRECTIONAL, false, false, 3, 0 );
    fill( TEXT_COLOR );
    text( "bidirectional", x + GENE_W/2 + 3, y );

    x = _legend_origin_x + 20;
    y += LEGEND_ELEMENT_OFFSET;

    textFont(_plotFont_8b);
    textAlign( LEFT, TOP );
    text( "Metabolites", x, y );

    x += 95;
    text( "Metrics", x, y );

    x = _legend_origin_x+GENE_W/2 + 20; 
    y += LEGEND_ELEMENT_OFFSET+4;

    strokeWeight( 1 );
    stroke( METABOLITE_COLOR );
    line( x, y, x+15, y );

    x = _legend_origin_x + 95 + 20 +GENE_W/2;
    y -= LEGEND_ELEMENT_OFFSET;

    // render the metric legend
    textAlign( LEFT, CENTER );
    for ( int i = 0; i < _pathway.numCorrelation(); i++ )
    {
      y +=  LEGEND_ELEMENT_OFFSET;
      
      if ( _correlation_rolled_over && (_rolled_over_correlation == i) )
      {
        renderGene( x, y, LEGEND, false, true, i, 0 );
        textFont(_plotFont_8b);
        fill( PHYLO_TREE_COLOR_ROLLED_OVER );
      }
      else
      {
        renderGene( x, y, LEGEND, false, false, i, 0 );
        textFont(_plotFont_8);
        fill( TEXT_COLOR );
      }
      
      text( _pathway.correlationName(i), x + GENE_W/2 + 5, y );
    }
    textFont(_plotFont_8);
  }

  int selectedNode( int index )
  {
    return _selected_node[index];
  }

  int numNodesSelected()
  { 
    return _num_nodes_selected;
  }

  boolean isSelected( int index )
  {
    for ( int i = 0; i < _num_nodes_selected; i++ )
      if ( _selected_node[i] == index ) return true;
    return false;
  }

  // function for dealing with mouseOver callbacks
  void overView( int x, int y )
  {
    // reset the rolled-over flags
    _node_rolled_over = false;

    // is the mouse over a node?
    int[] rv = overNode( x, y );
    _rolled_over_node = rv[0];
    _rolled_over_class = rv[1];
    if ( _rolled_over_node != -1 )
    {
      _node_rolled_over = true;
      _correlation_rolled_over = true;
      _rolled_over_correlation = _rolled_over_class;
    }
    
    // is the mouse over the metrics legend?
    else
    {
      overMetricLegend( x, y );
    }
  }

  // if a node is selected via the mouse then only one selection active
  void pressedInView( int x, int y, boolean ctrl )
  {   
    // is the mouse over the scrollbar?
    if ( _use_scroll ) _scrollbar_pressed = overScrollBar( x, y );

    if ( !_scrollbar_pressed )
    {
      // is the mouse over a node?
      int[] rv = overNodeForSelection( x, y );
      int code = rv[0];

      if ( code != -1 ) 
      {
        if ( !ctrl )
        {
          _selected_node[0] = code;
          _num_nodes_selected = 1;
          _remove_node[0] = false;
        }
        else
        {
          if ( _num_nodes_selected < MAX_NUM_SELECTION )
          {
            _selected_node[_num_nodes_selected] = code;
            _remove_node[_num_nodes_selected] = false;
            ++_num_nodes_selected;             
          }
        }
        setParalogFlags();
      }
    }
  }

  boolean overMetricLegend( int mx, int my )
  {
    _correlation_rolled_over = false;
    
    int x = _legend_origin_x + 120 +GENE_W;
    int y = _legend_origin_y-LEGEND_ELEMENT_OFFSET/4+2*LEGEND_ELEMENT_OFFSET+8;

    // render the metric legend
    textFont(_plotFont_8);
    for ( int i = 0; i < _pathway.numCorrelation(); i++ )
    {
      y +=  LEGEND_ELEMENT_OFFSET;
      
      if ( (mx >= x) && (mx <= x+textWidth(_pathway.correlationName(i))) && (my >= y-5) && (my <= y+5) )
      {
        _correlation_rolled_over = true;
        _rolled_over_correlation = i;
        
        return true;
      }
    }
    return false;
  }
  
  boolean overScrollBar( int x, int y )
  {
    if ( (x >= _scrollbar_x) && (x <= _scrollbar_x+SCROLLBAR_W) )
      if ( (y >= _pathway_start_y+_scroll_offset) && (y <= _pathway_start_y+_scroll_offset+_scroll_length) )
      {
        _scrollbar_pressed_y = y;
        return true;
      }
    return false;
  }

  void released()
  {
    _scrollbar_pressed = false;
  }

  void dragged( int x, int y )
  {
    if ( _scrollbar_pressed )
    {
      _scroll_offset += (y-_scrollbar_pressed_y);
      _scroll_offset = min( max( _scroll_offset, 0 ), _scrollbar_length-_scroll_length );
      computePathwayOffset();

      _scrollbar_pressed_y = y;
    }
  }

  void upKeyPressed()
  {  
    do
    {
      _selected_node[_num_nodes_selected-1] = (_selected_node[_num_nodes_selected-1] + (_pathway.numNodes()-1)) % _pathway.numNodes();
    } 
    while ( !_pathway.getNode(_selected_node[_num_nodes_selected-1]).haveCorrelation(0) );
  }

  void downKeyPressed()
  {  
    do
    {
      _selected_node[_num_nodes_selected-1] = (_selected_node[_num_nodes_selected-1] + 1) % _pathway.numNodes();
    } 
    while ( !_pathway.getNode(_selected_node[_num_nodes_selected-1]).haveCorrelation(0) );
  }

  int[] overNode( int x, int y )
  {
    if ( (y >= _pathway_start_y) && (y <= _pathway_stop_y) ) 
    {
      // check to see if the mouse is over any nodes
      int node_x, metab_y;
      int node_y = _pathway_origin_y-_pathway_offset;
      int node_center_y;
      PathwayNode node;
      int c;
      for ( int j = 0; j < _num_corr_selected; j++ )
      {
        if ( _num_corr_selected == 1 ) c = _single_selected_corr;
        else c = j;
        for ( int i = 0; i < _pathway.numNodes(); i++ )
        {
          node = _pathway.getNode(i);
          node_y = node.y(c)-_pathway_offset;

          if ( node.type() == GENE )
          {
            node_center_y = node_y+GENE_H/2;

            if ( node.haveCorrelation(c) && (y >= node_center_y-(GENE_H/2+OVER_TOLERANCE)) && (y <= node_center_y+(GENE_H/2+OVER_TOLERANCE)) )
            {
              node_x = node.x(c);  

              if ( (x >= node_x-(GENE_H/2+OVER_TOLERANCE)) && (x <= node_x+(GENE_H/2+OVER_TOLERANCE)) )
              {
                int[] rv = new int[2];
                rv[0] = i; // which node
                rv[1] = c; // which correlation
                return rv;
              }
            }
          }
          else 
          {
            node_center_y = node_y+METABOLITE_H/2;

            if ( node.haveCorrelation(c) && (y >= node_center_y-(METABOLITE_H+OVER_TOLERANCE)) && (y <= node_center_y+(METABOLITE_H+OVER_TOLERANCE)) )
            {
              node_x = node.x(c);  

              if ( (x >= _pathway_origin_x) && (x <= node_x+OVER_TOLERANCE) )
              {
                int[] rv = new int[2];
                rv[0] = i;
                rv[1] = c;
                return rv;
              }
            }
          }
        }
      }
    }

    int[] rv = new int[2];
    rv[0] = -1;
    rv[1] = -1;
    return rv;
  }

  int[] overNodeForSelection( int x, int y )
  {
    if ( (y >= _pathway_start_y) && (y <= _pathway_stop_y) ) 
    {

      // check to see if the mouse is over any nodes
      int node_x, metab_y;
      int node_y = _pathway_origin_y-_pathway_offset;
      int node_center_y;
      PathwayNode node;
      int c;
      for ( int j = 0; j < _num_corr_selected; j++ )
      {
        if ( _num_corr_selected == 1 ) c = _single_selected_corr;
        else c = j;
        for ( int i = 0; i < _pathway.numNodes(); i++ )
        {
          node = _pathway.getNode(i);
          node_y = node.y(c)-_pathway_offset;

          if ( node.type() == GENE )
          {
            if ( _num_corr_selected == 1 )
            {
              node_center_y = node_y+GENE_H/2;

              if ( node.haveCorrelation(c) && (y >= node_center_y-(GENE_H/2+OVER_TOLERANCE)) && (y <= node_center_y+(GENE_H/2+OVER_TOLERANCE)) )
              {
                node_x = node.x(c);  

                if ( (x >= node_x-(GENE_H/2+OVER_TOLERANCE)) && (x <= node_x+(GENE_H/2+OVER_TOLERANCE)) )
                {
                  /*
                  print( node.name() );
                   for ( int p = 0; p < _species.length; p++ ) print( "  " + node.paralogName(p) );
                   println();*/

                  int[] rv = new int[2];
                  rv[0] = i;
                  rv[1] = c;
                  return rv;
                }
              }
            }
            else
            {
              node_center_y = node_y+GENE_H/2;

              if ( node.haveCorrelation(c) && (y >= node_center_y-(GENE_H/2+OVER_TOLERANCE)) && (y <= node_center_y+(GENE_H/2+OVER_TOLERANCE)) )
              {
                node_x = min(node.x(0), node.x(1));
                int node_x2 = max(node.x(0), node.x(1));  

                if ( (x >= node_x-(GENE_H/2+OVER_TOLERANCE)) && (x <= node_x2+(GENE_H/2+OVER_TOLERANCE)) )
                {
                  int[] rv = new int[2];
                  rv[0] = i;
                  rv[1] = c;
                  return rv;
                }
              }
            }
          }
          else 
          {
            node_center_y = node_y+METABOLITE_H/2;

            if ( node.haveCorrelation(c) && (y >= node_center_y-(METABOLITE_H+OVER_TOLERANCE)) && (y <= node_center_y+(METABOLITE_H+OVER_TOLERANCE)) )
            {
              node_x = node.x(c);  

              if ( (x >= _pathway_origin_x) && (x <= node_x+OVER_TOLERANCE) )
              {
                int[] rv = new int[2];
                rv[0] = i;
                rv[1] = c;
                return rv;
              }
            }
          }
        }
      }
    }

    int[] rv = new int[2];
    rv[0] = -1;
    rv[1] = -1;
    return rv;
  }
}






































































