public static final int MAX_NUM_CORRELATIONS = 4;

String readConfigFile( String config_file )
{
  String[] rows = loadStrings(config_file);

  create_species( rows[0] );
  create_timepoints( rows[1] );

  // read the pathways
  _pathway = new Pathway();
  int counter = 0;
  String[] pathway_rows = loadStrings(splitTokens(rows[2])[1]);
  while ( counter < pathway_rows.length )
  {
    if ( pathway_rows[counter].length() > 0 )
      readPathwayFile( pathway_rows[counter] );
    ++counter;
  }

  // read the expression profiles
  counter = 0; 
  pathway_rows = loadStrings(splitTokens(rows[3])[1]);
  while ( counter < pathway_rows.length )
  {
    if ( pathway_rows[counter].length() > 0 )
      readExpressionFile( pathway_rows[counter] );
    ++counter;
  }

  // paralog file, if we have one
  int row_count = 4;
  pathway_rows =  splitTokens(rows[4]);
  if ( pathway_rows[0].equalsIgnoreCase("PARALOG_CONFIG_FILE:") )
  {
    pathway_rows = loadStrings(pathway_rows[1]);
    readParalogFile( splitTokens(pathway_rows[0])[1]);
    row_count = 5;
  }

  // read the correlation scores
  int corr_counter = 0;
  for ( int i = row_count; i < rows.length; i++ )
  {
    if ( rows[i].length() == 0 ) continue;

    pathway_rows = loadStrings(splitTokens(rows[i])[1]);
    String c_name = splitTokens(pathway_rows[0])[1];
    float c_min = parseFloat(splitTokens(pathway_rows[1])[1]);
    float c_max = parseFloat(splitTokens(pathway_rows[2])[1]);
    counter = 3; 
    while ( counter < pathway_rows.length )
    {
      if ( (pathway_rows[counter].length() > 0) && splitTokens(pathway_rows[counter])[0].equalsIgnoreCase("CORRELATION_FILE:") )
        readCorrelationFile( splitTokens(pathway_rows[counter])[1], c_name, corr_counter, c_min, c_max );
      else if ( (pathway_rows[counter].length() > 0) && splitTokens(pathway_rows[counter])[0].equalsIgnoreCase("SPECIES:") )
      {
        String[] s = splitTokens(pathway_rows[counter]);
        for ( int k = 1; k < s.length; k++)
          _pathway._correlation_species[corr_counter].add( s[k] );
      }
      ++counter;
    }

    ++corr_counter;
  }
  _pathway.numCorrelation( corr_counter );

  println( "number of correlation scores: " + corr_counter );

  // returnt the name of phylogentic tree file for creating the virtual tree
  return splitTokens(rows[0])[1];
}

void create_species( String species_file )
{
  String[] rows = loadStrings(splitTokens(species_file)[1]);
  String[] cols = splitTokens( rows[0] );
  _species = new String[cols.length];
  for ( int i = 0; i < _species.length; i++ ) 
    _species[i] = cols[i];
  print( _species.length + " species: " );
  for ( int i = 0; i < _species.length; i++ ) print( _species[i] + " " );
  println();
}

void create_timepoints( String timepoints_list )
{
  String[] cols = splitTokens( timepoints_list );
  _timepoints = new String[parseInt(cols[1])];
  for ( int i = 0; i < _timepoints.length; i++ ) 
    _timepoints[i] = "";
  print( _timepoints.length + " timepoints: " );
  for ( int i = 0; i < _timepoints.length; i++ ) print( _timepoints[i] + " " );
  println();
}

void readPathwayFile( String pathway_config )
{
  String[] config_cols = splitTokens( pathway_config );

  // set the start for this pathway and the stop for the previous
  if ( config_cols.length == 3 ) _pathway.setPathway( config_cols[2], _pathway.numNodes() );
  else _pathway.setPathway( _pathway.numNodes() );

  String[] rows = loadStrings( config_cols[1] );
  String[] cols, cols2;

  // check if the beginning of this pathway exists in the previous
  int index = 0;
  if (_pathway.numPathways() > 1)
  {
    // if this is a branch or a cycle, automatically make no this a discontinous pathway
    if ( !rows[0].substring(0,1).equalsIgnoreCase("B") && !rows[0].substring(0,1).equalsIgnoreCase("C") ) 
    {
      // work our way back up the previous pathway until (or not) we get a match
      APathway ap = _pathway.getPathway(_pathway.numPathways()-2);
      int counter = 0;
      cols = splitTokens( rows[0] );

      // find a match for the first node in the new pathway
      while ( (counter <= (ap._stop-ap._start)) && !_pathway.getNode(ap._stop - counter).name().equalsIgnoreCase(cols[1])  ) 
        ++counter;

      // if counter is beyond start, then there is no match. also check if the new pathway has
      //  enough nodes to be matched
      boolean exists = false;
      if ( (counter <= (ap._stop-ap._start)) && ((counter+1) <= rows.length) )
      {
        exists = true; 
        // check if the whole section is a match
        for ( int i = 1; i <= counter; i++ )
        {
          if ( rows[i].substring(0,1).equalsIgnoreCase("B") || rows[i].substring(0,1).equalsIgnoreCase("C") ) exists = false;
          else if ( !_pathway.getNode(ap._stop-counter+i).name().equalsIgnoreCase(splitTokens(rows[i])[1]) ) exists = false;
        }
      }

      if ( exists ) 
      {
        index = counter+1;
        _pathway.setPathwayStart( (ap._stop - counter), _pathway.numPathways()-1 );
      }
      else _pathway.setBreakPoint(_pathway.getPathway(_pathway.numPathways()-2)._stop); 
    }
    else _pathway.setBreakPoint(_pathway.getPathway(_pathway.numPathways()-2)._stop); 
  }

  // create each node of the pathway
  PathwayNode node;
  int node_counter = 0;
  Stack lifo = new Stack();
  Stack clifo = new Stack();
  int s_level = 0;
  char letter = 'a';
  Branch b;
  Cycle c;
  boolean render_next_name = false;
  boolean next_node_cycle = false;
  for ( int i = index; i < rows.length; i++ )
  {
    node = null;
    cols = splitTokens( rows[i] );

    // check for a branch point    
    if ( cols[0].equalsIgnoreCase("B1") )
    {
      // start of first branch is the next element to be added to the pathway
      lifo.push( new Branch( _pathway.numNodes(), s_level ) );
      b = (Branch)(lifo.peek());
      b.branchLetters( Character.toString(letter), Character.toString(++letter) );
      ++letter;
      if ( cols.length == 2 ) b.branch1Name( cols[1] );
      ++s_level;

      node = null;

      // for rendering breaks....
      if ( _pathway.numNodes() > 0 )
      {
        _pathway.setBranchPoint(_pathway.numNodes()-1);
        _pathway.getNode(_pathway.numNodes()-1).renderName(true);
      }
    }

    else if ( cols[0].equalsIgnoreCase("B2") )
    {
      // start of second branch is the next element to be added to the pathway
      b = (Branch)(lifo.peek());
      b.start2( _pathway.numNodes() );
      if ( cols.length == 2 ) b.branch2Name( cols[1] );

      node = null;

      // for rendering breaks....
      if ( _pathway.numNodes() > 0 )
        _pathway.setBranchPoint(_pathway.numNodes()-1);
    }

    else if ( cols[0].equalsIgnoreCase("BEND") )
    { 
      // stop of the branches is the last element added to the pathway
      b = (Branch)(lifo.peek());
      b.stop( _pathway.numNodes()-1 );
      _pathway.setBranch( (Branch)lifo.pop() );
      --s_level;

      node = null;

      // for rendering breaks....
      if ( _pathway.numNodes() > 0 )
      {
        _pathway.setBranchPoint(_pathway.numNodes()-1);
        render_next_name = true;
      }
    }

    else if ( cols[0].equalsIgnoreCase("C") )
    { 
      // start of a cycle is the next element to be added to the pathway
      clifo.push( new Cycle( _pathway.numNodes() ) );
      c = (Cycle)(clifo.peek());
      if ( cols.length == 2 ) c.cycleName( cols[1] );

      node = null;
      render_next_name = true;
      next_node_cycle = true;
    }

    else if ( cols[0].equalsIgnoreCase("CEND") )
    { 
      // stop of the branches is the last element added to the pathway
      c = (Cycle)(clifo.peek());
      c.stop( _pathway.numNodes()-1 );
      _pathway.setCycle( (Cycle)clifo.pop() );

      node = null;
      _pathway.getNode(_pathway.numNodes()-1).renderName(true);
      _pathway.getNode(_pathway.numNodes()-1).cyclePoint(true);
    }

    // check for a metabolite
    else if ( cols[0].equalsIgnoreCase("M") )
      node = new PathwayNode( cols[1] );

    // check for a forward gene
    else if ( cols[0].equalsIgnoreCase("GF") )
      node = new PathwayNode( cols[1], FORWARD );

    // check for a reverse gene
    else if ( cols[0].equalsIgnoreCase("GR") )
      node = new PathwayNode( cols[1], REVERSE );

    // check for a bidirectional gene
    else if ( cols[0].equalsIgnoreCase("GFR") )
      node = new PathwayNode( cols[1], BIDIRECTIONAL );

    // if the node is something else, we have a problem
    else
    {
      println( cols[0] + " is not a valid node type!" );
      node = null;
    }

    // check for a shortened name
    if ( (cols.length == 3) && (node != null) ) node.shortName( cols[2] );

    // add the node to the pathway
    if ( node != null )
    {
      if ( render_next_name )
      {
        node.renderName(true);
        render_next_name = false;
      }
      if ( next_node_cycle )
      {
        node.cyclePoint( true );
        next_node_cycle = false;
      }
      _pathway.setNode( node );
      ++node_counter;
    }
  }

  // set the stop point of the pathway
  _pathway.setPathwayStop( _pathway.numNodes()-1, _pathway.numPathways()-1 );

  // tell the start and stop nodes to render their names
  _pathway.getNode(_pathway.getPathway(_pathway.numPathways()-1)._start).renderName( true );
  _pathway.getNode(_pathway.getPathway(_pathway.numPathways()-1)._stop).renderName( true );
}

void readExpressionFile( String exp_file )
{
  String[] cols = splitTokens( exp_file );
  String[] rows = loadStrings( cols[1] );

  // for each element in the profile file, search the pathway nodes for a 
  //  matching name (TODO: something more efficient!)
  PathwayNode node;
  float[] profile = new float[_species.length*_timepoints.length];

  for ( int i = 0; i < rows.length; i++ )
  {
    cols = splitTokens( rows[i] );

    // check that there are the right number of expression values
    if ( cols.length != (_species.length*_timepoints.length+1) )
    {
      println( cols[0] + ": WRONG NUMBER OF EXPRESSION ELEMENTS!!!" );
      continue;
    }

    boolean found_node = false;
    for ( int j = 0; j < _pathway.numNodes(); j++ )
    {
      node = _pathway.getNode(j);
      if ( (node.name()).equalsIgnoreCase(cols[0]) )
      {   
        found_node = true;

        if ( !node.haveExpressionProfile() )
        {  
          for ( int k = 0; k < _species.length*_timepoints.length; k++ )
          {
            // check for NaNs
            if ( cols[k+1].equalsIgnoreCase( "NaN" ) ) profile[k] = Float.NaN;
            else profile[k] = parseFloat(cols[k+1]);
          }

          node.createExpressionProfile( profile );
        }
      }
      /*
      if ( (j == (_pathway.numNodes()-1)) && !found_node )
       println( "Didn't find " + rows[i] );*/
    } 
  }
}

void readCorrelationFile( String corr_file, String corr_name, int which_corr, float corr_min, float corr_max )
{
  String[] cols;
  String[] rows = loadStrings( corr_file );

  // for each element in the correlation file, search the pathway nodes for a 
  //  matching name (TODO: something more efficient!)
  PathwayNode node;
  for ( int i = 0; i < rows.length; i++ )
  {
    cols = splitTokens( rows[i] );

    boolean found_node = false;
    for ( int j = 0; j < _pathway.numNodes(); j++ )
    {
      node = _pathway.getNode(j);
      if ( (node.name()).equalsIgnoreCase(cols[0]) )
      {
        found_node = true;
        node.correlation( parseFloat(cols[1]), which_corr );
      }
      /*   
       if ( (j == (_pathway.numNodes()-1)) && !found_node )
       println( "Didn't find " + rows[i] );*/
    }
  }

  // parse the min and max correlation values from the config file
  _pathway.correlationMinMax( corr_min, corr_max );
  _pathway.correlationName( corr_name, which_corr );
}

void readParalogFile( String file )
{
  String[] rows = loadStrings( file );

  // go through the header and get the indices of the species
  String[] cols = splitTokens( rows[0] );
  int[] indices = new int[_species.length];
  for ( int i = 0; i < _species.length; i++ )
  {
    for ( int j = 1; j < cols.length; j++ )
    {
      if ( _species[i].equalsIgnoreCase(cols[j] ) ) 
        indices[i] = j;
    }
  }

  // got through the pathway nodes and set the paralog names
  String[] names = new String[_species.length];
  String gene;
  for ( int i = 1; i < rows.length; i++ )
  {
    cols = splitTokens( rows[i] );
    gene = cols[0];
    for ( int j = 0; j < _pathway.numNodes(); j++ )
    {
      if ( _pathway.getNode(j).name().equals(gene) )
      {
        for ( int k = 0; k < _species.length; k++ )
          names[k] = cols[indices[k]];

        _pathway.getNode(j).paralogNames( names );
        
        break;
      } 
    }
  }
}











