public final color UNSELECTED_PLOT_LINE_COLOR = color( 100 );
public final color ZERO_LINE_COLOR = color( 200 );
public final color PLOT_LINE_COLOR = color( 50 );
public final color PLOT_FILL_COLOR = color( 200 );
public final color PLOT_PARALOG_FILL_COLOR = color( 206, 215, 222 );
public final color PLOT_OUTLINE_COLOR = color( 0 );
public final color PHYLO_TREE_COLOR = color( 150 );
public final color PHYLO_TREE_COLOR_ROLLED_OVER = color( 153, 0, 0 );
public final static int INDIVIDUAL_PLOT_SPACING = 8;
public final static int PLOT_SPACING = 40;
public final static int COLUMN_SPACING = INDIVIDUAL_PLOT_SPACING;
public final static int NORMALIZATION_BOX_WIDTH = 12;
public final static int TREE_OFFSET = 40;
public final static int OVERLAY_OFFSET = 20;
public final static int REMOVE_BOX_OFFSET = 30;
public final static int REMOVE_BOX_WIDTH = 8;

public final static int MAX_INDIVIDUAL_PLOT_WIDTH = 60;

boolean[] _species_rolled_over;
boolean[] _selected_node_rolled_over;

int _tree_center_x, _shapemaps_x, _overview_center_x;

boolean _timeseries_rolled_over;
int[] _rolled_over_timeseries;

boolean[] _remove_node;

class TimeSeriesView
{
  int _width, _height;
  int _individual_plot_width, _individual_plot_height, _individual_timestep_width;
  int _combined_plot_width, _combined_plot_height, _combined_timestep_width;
  int[] _individual_plot_origin;
  int[] _combined_plot_origin, _normalization_option_box_1, _normalization_option_box_2;
  int _tree_width, _tree_interlevel_width;

  boolean _timeseries_normalized;

  PhyTreeNode _phy_tree;
  String _phy_tree_file;

  int _individual_plot_combined_height;

  TimeSeriesView( String phylo_tree_file )
  {
    _individual_plot_origin = new int[2];
    _combined_plot_origin = new int[2];
    _normalization_option_box_1 = new int[2];
    _normalization_option_box_2 = new int[2];

    _tree_width = 75;
    _individual_plot_origin[0] = _tree_width + TREE_OFFSET;
    _individual_plot_origin[1] = REMOVE_BOX_OFFSET-5;

    _tree_center_x = _individual_plot_origin[0]/2;
    _shapemaps_x = _individual_plot_origin[0];

    _phy_tree_file = phylo_tree_file;

    _timeseries_normalized = true;
    _species_rolled_over = new boolean[_species.length];
    for ( int i = 0; i < _species_rolled_over.length; i++ ) _species_rolled_over[i] = false;
    _selected_node_rolled_over = new boolean[UPPER_BOUND_MAX_NUM_SELECTION];
    for ( int i = 0; i < _selected_node_rolled_over.length; i++ ) _selected_node_rolled_over[i] = false;

    _timeseries_rolled_over = false;
    _rolled_over_timeseries = new int[2];
    _rolled_over_timeseries[0] = 0;
    _rolled_over_timeseries[1] = 0;

    _remove_node = new boolean[MAX_NUM_SELECTION];
    for ( int i = 0; i < _remove_node.length; i++ ) _remove_node[i] = false;
  }

  void setDimensions( int w, int h )
  {
    _width = w;
    _height = h;

    _individual_plot_height = (_height - _individual_plot_origin[1] - OVERLAY_OFFSET - INDIVIDUAL_PLOT_SPACING*(_species.length))/(_species.length+1);
    _individual_plot_height = min( _individual_plot_height, MAX_INDIVIDUAL_PLOT_WIDTH );
    _individual_plot_width = _individual_plot_height;
    MAX_NUM_SELECTION = max( 1, floor( (_width - _tree_width - TREE_OFFSET - OVERLAY_OFFSET) / (_individual_plot_width+COLUMN_SPACING) - 1.5) );
    _num_nodes_selected = min( _num_nodes_selected, MAX_NUM_SELECTION );
    _individual_timestep_width = _individual_plot_width / (_timepoints.length-1);
    _individual_plot_combined_height = _individual_plot_height*_species.length + INDIVIDUAL_PLOT_SPACING*(_species.length-1);

    boolean[] tmp = new boolean[_remove_node.length];
    int i;
    for ( i = 0; i < tmp.length; i++ ) tmp[i] = _remove_node[i];
    _remove_node = new boolean[MAX_NUM_SELECTION];
    for ( i = 0; i < min(_remove_node.length,tmp.length); i++ ) _remove_node[i] = tmp[i];
    for ( i = i; i < _remove_node.length; i++ ) _remove_node[i] = false;

    _combined_plot_width = _individual_plot_width;
    _combined_plot_height = _individual_plot_height;    
    _combined_timestep_width = _combined_plot_width / (_timepoints.length-1);
    _combined_plot_origin[0] = _individual_plot_origin[0];
    _combined_plot_origin[1] = _individual_plot_origin[1] + (_species.length)*_individual_plot_height + INDIVIDUAL_PLOT_SPACING*(_species.length) + _combined_plot_height + OVERLAY_OFFSET;

    _normalization_option_box_1[0] = _normalization_option_box_2[0] = _combined_plot_origin[0] + _combined_plot_width - NORMALIZATION_BOX_WIDTH;
    _normalization_option_box_1[1] = _combined_plot_origin[1] + PLOT_SPACING + NORMALIZATION_BOX_WIDTH; 
    _normalization_option_box_2[1] = _normalization_option_box_1[1] + NORMALIZATION_BOX_WIDTH * 3 / 2; 

    // create the tree last because we need to know the 
    //  positions of the individual species plots
    createPhylogeneticTree( _phy_tree_file );

    setParalogFlags();
  }

  void pressedInView( int mx, int my )
  {
    int y = _individual_plot_origin[1] - REMOVE_BOX_OFFSET;
    if ( abs(my-y) > REMOVE_BOX_WIDTH/2 ) return;

    int x = _individual_plot_origin[0] + _individual_plot_width/2;    
    for ( int i = 0; i < _num_nodes_selected; i++ )
    {
      if ( abs(mx-x) <= REMOVE_BOX_WIDTH/2 )
      {
        _remove_node[i] = !_remove_node[i];
        return;
      }
      x += (_individual_plot_width + COLUMN_SPACING);
    }

    x += OVERLAY_OFFSET;
    textFont( _plotFont_8 );
    if ( abs(mx-x) <= (textWidth("remove")/2+2) )
    {
      // store the selected node array
      int[] tmp_sn = new int[_num_nodes_selected];
      for ( int i = 0; i < tmp_sn.length; i++ ) tmp_sn[i] = _selected_node[i];

      // remove the selected nodes from the list
      int count = 0;
      for ( int i = 0; i < _num_nodes_selected; i++ )
      {
        if ( !_remove_node[i] )
        {
          _selected_node[count] = tmp_sn[i];
          ++count;
        }
        else _remove_node[i] = false;
      }
      _num_nodes_selected = count;
    }
  }

  void rolledOver( int x, int y )
  {
    for ( int i = 0; i < _species_rolled_over.length; i++ ) _species_rolled_over[i] = false;
    for ( int i = 0; i < _selected_node_rolled_over.length; i++ ) _selected_node_rolled_over[i] = false;

    _timeseries_rolled_over = false;

    // check for roll over in the tree
    //if ( x < _tree_width )
    // go through the tree and look for something rolled over
    _phy_tree.checkForRollOver(x,y);
    /*else*/    if ( (x < _individual_plot_origin[0]) && (x > _tree_width) )
      // look for a rolled over name
      speciesNameRolledOver( y );
    else if ( (x > _individual_plot_origin[0]) && (y <= _individual_plot_origin[1]) && (y >= (_individual_plot_origin[1]-OVER_TOLERANCE*20)) )
      nodeNameRolledOver( x );
    else if ( (y > _individual_plot_origin[1]) && (y < (_individual_plot_origin[1]+_individual_plot_combined_height)) )
      timeseriesRolledOver( x, y );
  }

  void timeseriesRolledOver( int mx, int my )
  {
    // the top left corner of the first time plot    
    int x;
    int y = _individual_plot_origin[1];


    for ( int i = 0; i < _species.length; i++ )
    {
      x = _individual_plot_origin[0];
      for ( int j = 0; j < _num_nodes_selected; j++ )
      {
        if ( _paralogs[i][j] && (mx >= x) && (mx <= (x+_individual_plot_width)) && (my >= y) && (my <= (y+_individual_plot_height)) )
        {          
          _timeseries_rolled_over = true;
          _rolled_over_timeseries[0] = i;
          _rolled_over_timeseries[1] = j;
          return;
        }        
        x += _individual_plot_width + COLUMN_SPACING;
      }
      y += _individual_plot_height+INDIVIDUAL_PLOT_SPACING;
    }
  }

  void speciesNameRolledOver( int y )
  {
    int node_y = _individual_plot_origin[1] + _individual_plot_height/2;
    for ( int i = 0; i < _species.length; i++ )
    {
      if ( abs(node_y-y) <= OVER_TOLERANCE*10 ) 
      {
        _species_rolled_over[i] = true;
        return;
      }

      node_y += _individual_plot_height+INDIVIDUAL_PLOT_SPACING;
    }
  }

  void nodeNameRolledOver( int x )
  {
    int node_x = _individual_plot_origin[0];
    for ( int i = 0; i < _num_nodes_selected; i++ )
    {
      if ( (x-node_x) <= _individual_plot_width ) 
      {
        _selected_node_rolled_over[i] = true;
        return;
      }

      node_x += _individual_plot_width + COLUMN_SPACING;
    }
  }

  void render()
  {
    textFont(_plotFont_8); 
    for ( int i = 0; i < _num_nodes_selected; i++ )
    {
      renderIndividualTimeSeries(i);
      renderNodeCombinedTimeSeries(i);
    }

    for ( int i = 0; i < _num_nodes_selected; i++ )
      if ( _selected_node_rolled_over[i] ) 
      {
        renderRolledOverNodeName(i);
        break;
      }

    _overview_center_x = -1;
    renderSpeciesCombinedTimeSeries();

    //renderNormalizationOption();
    renderPhylogeneticTree();
  }

  void renderRolledOverNodeName( int index )
  {
    PathwayNode node = _pathway.getNode(_selected_node[index]);

    textFont(_plotFont_8); 

    int box_h = TEXT_HEIGHT+2;
    int box_w;
    box_w = AVE_TEXT_WIDTH*(node.name().length()+2);

    int mx = mouseX - _timeseries_view_origin[0] + 5;
    int my = mouseY - _timeseries_view_origin[1] - 5;

    noStroke();
    //fill( PATHWAY_COLOR, 200 );
    fill( color(255), 180 );
    rectMode( CORNER );
    rect( mx, my, box_w, -box_h );

    textAlign( LEFT, BOTTOM );
    fill( color(0) );   
    text(node.name(),  mx+2, my+1 );
  }

  void renderIndividualTimeSeries( int index )
  {
    // the bottom left corner of the first time plot    
    int x = _individual_plot_origin[0] + index*(_individual_plot_width + COLUMN_SPACING);
    int y = _individual_plot_origin[1] + _individual_plot_height;

    // write the name of the selected node
    textFont(_plotFont_12); 
    textAlign( LEFT, BOTTOM );
    if ( _selected_node_rolled_over[index] )
      fill( PHYLO_TREE_COLOR_ROLLED_OVER );
    else fill( TEXT_COLOR );
    text( _pathway.getNode(_selected_node[index]).shortName(), x, y-_individual_plot_height-5 );

    // draw the remove box
    stroke( TEXT_COLOR );
    strokeWeight( 1 );
    if ( _remove_node[index] ) fill( PHYLO_TREE_COLOR_ROLLED_OVER ); 
    else noFill();
    rectMode( CENTER );
    rect( x+_individual_plot_width/2, y-_individual_plot_height-REMOVE_BOX_OFFSET, REMOVE_BOX_WIDTH, -REMOVE_BOX_WIDTH );

    // render the plots 
    float min_v, max_v, v_diff;
    int zero_y;
    ExpressionProfile profile;
    for ( int i = 0; i < _species.length; i++ )
    {
      // render the background color
      rectMode( CORNER );
      noSmooth();
      noStroke();  
      fill( PATHWAY_COLOR );
      rect( x, y, _individual_plot_width, -_individual_plot_height ); 

      // render the time plot  
      profile = _pathway.getNode(_selected_node[index]).expressionProfile(i);
      min_v = profile.minValue();
      max_v = profile.maxValue();
      v_diff = max_v - min_v + 1.0e-6;
      //zero_y = y - round((0.0 - min_v) / v_diff*_individual_plot_height);

      int vert_x = x - _individual_timestep_width;
      int vert_y, vert_y_2;

      //
      // NOTE: this is a convoluted way to do this because of the NaNs!
      //
      noStroke();
      noSmooth();
      if ( _paralogs[i][index] ) fill( PLOT_PARALOG_FILL_COLOR );
      else fill( PLOT_FILL_COLOR );
      for ( int j = 0; j < _timepoints.length-1; j++ )
      {
        vert_x += _individual_timestep_width;

        //check if either of the points is a NaN
        if ( Float.isNaN(profile.value(j)) || Float.isNaN(profile.value(j+1)) ) 
          continue;

        // since the data exists at this point and the next, draw a filled quad
        beginShape(); 

        vert_y = y - round(((profile.value(j) - min_v) / v_diff)*_individual_plot_height); 
        vertex( vert_x, vert_y );

        vert_y = y - round(((profile.value(j+1) - min_v) / v_diff)*_individual_plot_height);          
        if ( j == _timepoints.length-2 )
        {
          vertex( x+_individual_plot_width, vert_y );          
          vertex( x+_individual_plot_width, y );
        }
        else
        {
          vertex( vert_x+_individual_timestep_width, vert_y );          
          vertex( vert_x+_individual_timestep_width, y );
        }
        vertex( vert_x, y );

        endShape( CLOSE );
      }

      // draw the zero line
      /*strokeWeight( 1 );
       stroke( ZERO_LINE_COLOR );
       noFill();
       line( x, zero_y, x+_individual_plot_width, zero_y);
       fill( TEXT_COLOR );
       textAlign( LEFT, CENTER );
       text( "0", x+_individual_plot_width+2, zero_y );*/

      // and draw the line plot
      noSmooth();
      vert_x = x - _individual_timestep_width;
      //stroke( PLOT_LINE_COLOR );
      stroke( color(150) );
      strokeWeight( 1 );
      smooth();
      for ( int j = 0; j < _timepoints.length-1; j++ )
      {
        vert_x += _individual_timestep_width;

        //check if either of the points is a NaN
        if ( Float.isNaN(profile.value(j)) || Float.isNaN(profile.value(j+1)) ) 
          continue;        

        vert_y = y - round(((profile.value(j) - min_v) / v_diff)*_individual_plot_height); 
        vert_y_2 = y - round(((profile.value(j+1) - min_v) / v_diff)*_individual_plot_height);   
        if ( j == _timepoints.length-2 )
          line( vert_x, vert_y, x+_individual_plot_width, vert_y_2 );
        else
          line( vert_x, vert_y, vert_x+_individual_timestep_width, vert_y_2 );
      }

      // outline the plots
      noSmooth();
      strokeWeight( 1 );  
      //stroke( PLOT_OUTLINE_COLOR );
      stroke( color(175) );
      noFill();
      rect( x, y, _individual_plot_width, -_individual_plot_height );   

      y += _individual_plot_height+INDIVIDUAL_PLOT_SPACING;
    }

    // check for rolled over
    if ( _timeseries_rolled_over )
    {
      PathwayNode node = _pathway.getNode(_selected_node[_rolled_over_timeseries[1]]);
      textFont(_plotFont_8); 

      int box_h = TEXT_HEIGHT+2;
      int box_w;
      box_w = AVE_TEXT_WIDTH*(node.paralogName(_rolled_over_timeseries[0]).length()+2);

      int mx = mouseX - _timeseries_view_origin[0] + 5;
      int my = mouseY - _timeseries_view_origin[1] - 5;

      noStroke();
      //fill( PATHWAY_COLOR, 200 );
      fill( color(255), 180 );
      rectMode( CORNER );
      rect( mx, my, box_w, -box_h );

      textAlign( LEFT, BOTTOM );
      fill( color(0) );   
      text(node.paralogName(_rolled_over_timeseries[0]),  mx+2, my+1 );
    }
  } 

  void renderNodeCombinedTimeSeries( int index )
  {
    // the bottom left corner of the combined time plot    
    int x = _combined_plot_origin[0] + index*(_individual_plot_width + COLUMN_SPACING);
    int y = _combined_plot_origin[1];  

    float min_v = _pathway.getNode(_selected_node[index]).minValueOverSpecies();
    float max_v = _pathway.getNode(_selected_node[index]).maxValueOverSpecies();

    // render the plots 
    rectMode( CORNER );



    // first, render all the not rolled over lines
    //smooth();
    noSmooth();
    strokeWeight( 1 );
    for ( int i = 0; i < _species.length; i++ )
      if ( !_species_rolled_over[i] ) renderLinePlot( index, i, x, y, /*PLOT_LINE_COLOR*/color(100), min_v, max_v );

    // then render all the rolled over lines
    strokeWeight( 1.5 );
    for ( int i = 0; i < _species.length; i++ )
      if ( _species_rolled_over[i] ) renderLinePlot( index, i, x, y, PHYLO_TREE_COLOR_ROLLED_OVER, min_v, max_v );

    // outline the plots
    noSmooth();
    strokeWeight( 1 );  
    stroke( /*PLOT_OUTLINE_COLOR*/ color(100) );
    noFill();
    rect( x, y, _combined_plot_width, -_combined_plot_height ); 

    // write the min and max values
    textFont(_plotFont_8); 
    fill( PLOT_LINE_COLOR );

    float rounded_v = round( min_v*10.0 )/10.0;
    String s = nfs( rounded_v, 1, 1 );
    textAlign( RIGHT, TOP );    
    text( s, x+_combined_plot_width, y );

    rounded_v = round( max_v*10.0 )/10.0;
    s = nfs( rounded_v, 1, 1 );
    textAlign( RIGHT, BOTTOM );    
    text( s, x+_combined_plot_width, y-_combined_plot_height );

    // render the zero line
    //    int zero_y = y + (int)((min_v/(max_v-min_v))*(float)_combined_plot_height);
    //    strokeWeight( 1 );  
    //    stroke( PHYLO_TREE_COLOR_ROLLED_OVER/*color(175)*/ );
    //    line( x, zero_y, x+_combined_plot_width, zero_y );
  }

  void renderSpeciesCombinedTimeSeries()
  {
    // the bottom left corner of the combined time plot    
    int x = _individual_plot_origin[0] + _num_nodes_selected*(_individual_plot_width + COLUMN_SPACING) + OVERLAY_OFFSET;
    int y = _individual_plot_origin[1] + _individual_plot_height; 

    // draw the remove box
    textFont(_plotFont_8); 

    stroke( TEXT_COLOR );
    strokeWeight( 1 );
    fill( color(240) );
    rectMode( CENTER );
    rect( x+_individual_plot_width/2-1, y-_individual_plot_height-REMOVE_BOX_OFFSET, textWidth("remove")+4, -REMOVE_BOX_WIDTH-2 );

    fill( TEXT_COLOR );
    textAlign( CENTER, CENTER );
    text( "remove", x+_individual_plot_width/2, y-_individual_plot_height-REMOVE_BOX_OFFSET-3 );

    _overview_center_x = x;

    // render the plots 
    int zero_y;
    rectMode( CORNER );
    for ( int i = 0; i < _species.length; i++ )
    {
      // get the min and max value over the nodes
      float min_v = Float.MAX_VALUE;
      float max_v = Float.MIN_VALUE;
      for ( int j = 0; j < _num_nodes_selected; j++ )  
      {
        min_v = min( min_v, _pathway.getNode(_selected_node[j]).expressionProfile(i).minValue() );
        max_v = max( max_v, _pathway.getNode(_selected_node[j]).expressionProfile(i).maxValue() );
      }
      if ( _num_nodes_selected == 0 ) min_v = max_v = 0.0;



      // render all the not-rolled-over lines
      noSmooth();
      strokeWeight( 1 );
      for ( int j = 0; j < _num_nodes_selected; j++ )  
        if ( !_selected_node_rolled_over[j] ) renderLinePlot( j, i, x, y, /*PLOT_LINE_COLOR*/color(100), min_v, max_v );

      // render the rolled-over node lines
      strokeWeight( 1.5 );
      for ( int j = 0; j < _num_nodes_selected; j++ )  
        if ( _selected_node_rolled_over[j] ) renderLinePlot( j, i, x, y, PHYLO_TREE_COLOR_ROLLED_OVER, min_v, max_v );

      // outline the plots
      noSmooth();
      strokeWeight( 1 );  
      stroke( /*PLOT_OUTLINE_COLOR*/color(100) );
      noFill();
      rect( x, y, _combined_plot_width, -_combined_plot_height ); 

      // write the min and max values
      textFont(_plotFont_8); 
      fill( PLOT_LINE_COLOR );

      float rounded_v = round( min_v*10.0 )/10.0;
      String s = nfs( rounded_v, 1, 1 );
      textAlign( LEFT, BOTTOM );    
      text( s, x+_combined_plot_width+2, y+3 );

      rounded_v = round( max_v*10.0 )/10.0;
      s = nfs( rounded_v, 1, 1 );
      textAlign( LEFT, TOP );    
      text( s, x+_combined_plot_width+2, y-_combined_plot_height-3 );

      // render the zero line
      //      zero_y = y + (int)((min_v/(max_v-min_v))*(float)_combined_plot_height);
      //      strokeWeight( 1 );  
      //      stroke( PHYLO_TREE_COLOR_ROLLED_OVER/*color(175)*/ );
      //      line( x, zero_y, x+_combined_plot_width, zero_y );

      y += _individual_plot_height+INDIVIDUAL_PLOT_SPACING;
    }
  }

  void renderLinePlot( int node_index, int profile_index, int x, int y, color c, float min_v, float max_v )
  {
    float v_diff = max_v - min_v + 1.0e-6;
    int vert_y, vert_y_0, vert_x;
    ExpressionProfile profile;
    noFill();
    stroke( c );
    for ( int i = 0; i < _species.length; i++ )
    {
      // render the time plot
      profile = _pathway.getNode(_selected_node[node_index]).expressionProfile(profile_index);

      vert_x = x - _combined_timestep_width;
      for ( int j = 0; j < _timepoints.length-1; j++ )
      {
        vert_x += _combined_timestep_width;

        //check if either of the points is a NaN
        if ( Float.isNaN(profile.value(j)) || Float.isNaN(profile.value(j+1)) ) 
          continue;

        // since the data exists at this point and the next, draw a line
        vert_y = y - round(((profile.value(j+1) - min_v) / v_diff)*_combined_plot_height);
        vert_y_0 = y - round(((profile.value(j) - min_v) / v_diff)*_combined_plot_height); 
        if ( j == _timepoints.length-2 )
          line( vert_x, vert_y_0, x+_combined_plot_width, vert_y );
        else
          line( vert_x, vert_y_0, vert_x+_combined_timestep_width, vert_y );
      }
    }
  }

  /*void renderNormalizationOption()
   {
   textAlign( RIGHT, BOTTOM ); 
   stroke( UNSELECTED_PLOT_LINE_COLOR );
   strokeWeight( 1 );
   
   fill( TEXT_COLOR );
   text( "normalized", _normalization_option_box_1[0]-5, _normalization_option_box_1[1] );
   if ( _timeseries_normalized ) fill( SELECTED_NODE_COLOR );
   else noFill();    
   rect( _normalization_option_box_1[0], _normalization_option_box_1[1], NORMALIZATION_BOX_WIDTH, -NORMALIZATION_BOX_WIDTH );
   
   fill( TEXT_COLOR );
   text( "absolute", _normalization_option_box_2[0]-5, _normalization_option_box_2[1] );
   if ( !_timeseries_normalized ) fill( SELECTED_NODE_COLOR );
   else noFill(); 
   rect( _normalization_option_box_2[0], _normalization_option_box_2[1], NORMALIZATION_BOX_WIDTH, -NORMALIZATION_BOX_WIDTH );
   
   }*/

  void renderPhylogeneticTree()
  {
    // first, just render the root node line
    stroke( PHYLO_TREE_COLOR );
    strokeWeight( 1 );
    int x = 0;
    int y = _phy_tree.y();
    noSmooth();
    line( x, y, _phy_tree.x(), y );

    _phy_tree.render();

    // write the name of the species
    x = _individual_plot_origin[0];
    y = _individual_plot_origin[1] + _individual_plot_height;
    for ( int i = 0; i < _species.length; i++ )
    {
      textAlign( RIGHT, CENTER );
      fill( TEXT_COLOR );
      textFont(_plotFont_12); 
      if ( _species_rolled_over[i] ) fill( PHYLO_TREE_COLOR_ROLLED_OVER );
      else if ( _correlation_rolled_over )
      {
        for ( int j = 0; j < _pathway._correlation_species[_rolled_over_correlation].size(); j++ )
        {
          if ( ((String)_pathway._correlation_species[_rolled_over_correlation].get(j)).equalsIgnoreCase(_species[i]) )
          {
            fill( PHYLO_TREE_COLOR_ROLLED_OVER );
            textFont(_plotFont_12); 
            break;
          }
        }
      }
      noStroke();
      text( _species[i], x-5, y-_individual_plot_height/2-2 );

      y += _individual_plot_height+INDIVIDUAL_PLOT_SPACING;
    }
  }

  void createPhylogeneticTree( String file )
  {
    String[] rows = loadStrings( file );
    String[] cols = splitTokens( rows[0] );

    // using the number of rows and number of elements in the first row
    //  create a temp table to compute into
    TempTableForPhyTree[][] table;
    table = new TempTableForPhyTree[rows.length][cols.length];

    // compute the width of each level of the tree
    _tree_interlevel_width = _tree_width / rows.length;

    // first, get the positions for the leaf nodes
    int x = _tree_width;
    int y = _individual_plot_origin[1] + _individual_plot_height/2;
    for ( int i = 0; i < cols.length; i++ )
    {
      table[0][i] = new TempTableForPhyTree( x, y, cols[i] );

      y += _individual_plot_height+INDIVIDUAL_PLOT_SPACING;
    }
    x -= _tree_interlevel_width;

    // compute the positions and fill up the table
    String[] indices_char;
    int[] indices = new int[4];
    for ( int i = 1; i < rows.length; i++ )
    {
      cols = splitTokens( rows[i] );
      for ( int j = 0; j < cols.length; j++ )
      {
        indices_char = splitTokens( cols[j], ";,[]" );
        indices[0] = parseInt(indices_char[0]);
        indices[1] = parseInt(indices_char[1]);
        indices[2] = parseInt(indices_char[2]);
        indices[3] = parseInt(indices_char[3]);
        y = table[indices[0]][indices[1]].y() + table[indices[2]][indices[3]].y();
        y /= 2;

        table[i][j] = new TempTableForPhyTree( x, y );
        table[i][j].childrenIndices( indices[0], indices[1], indices[2], indices[3] );
      }
      x -= _tree_interlevel_width;
    }

    // now, build the tree
    _phy_tree = new PhyTreeNode( rows.length-1, 0, table );
  }
}

/**************************************************************/
class PhyTreeNode
{
  int _x, _y; 
  PhyTreeNode[] _child;
  boolean _have_children;
  String _species_name; // this is for the leaf nodes
  int _species_index;
  boolean _rolled_over;

  PhyTreeNode( int i, int j, TempTableForPhyTree[][] table )
  {
    _x = table[i][j].x();
    _y = table[i][j].y();

    // check for children -- if this in NOT the 0th level, then yes!
    if ( i != 0 )
    {
      _have_children = true;
      _child = new PhyTreeNode[2];
      _child[0] = new PhyTreeNode( table[i][j]._child1[0], table[i][j]._child1[1], table );
      _child[1] = new PhyTreeNode( table[i][j]._child2[0], table[i][j]._child2[1], table );
    }
    else
    {
      _have_children = false;
      _species_name = table[i][j].name();

      // find the species index that this leaf node corresponds to
      boolean still_going = true;
      int s = 0;
      while ( still_going )
      {
        if ( _species_name.equalsIgnoreCase(_species[s]) )
        {
          _species_index = s;
          still_going = false;
        }
        else
        {
          ++s;
          if ( s == _species.length )
          {
            println( "NEVER FOUND SPECIES NAME in PhyTreeNode()" );
            still_going = false;
          }
        }
      }
    }

    _rolled_over = false;
  }

  int x()
  {
    return _x;
  }

  int y()
  { 
    return _y;
  }

  void rolledOver( boolean ro )
  {
    _rolled_over = ro;
    if ( _rolled_over )
    {
      if ( _have_children )
      {
        _child[0].rolledOver(true);
        _child[1].rolledOver(true);
      }
      else _species_rolled_over[_species_index] = true;
    }
  }

  void checkForRollOver( int x, int y )
  {
    if ( (abs(x-_x) <= OVER_TOLERANCE*10) && (abs(y-_y) <= OVER_TOLERANCE*10) )
      rolledOver(true);
    else
    {
      rolledOver(false);
      if ( _have_children )
      { 
        _child[0].checkForRollOver(x,y);
        _child[1].checkForRollOver(x,y);
      }
    }
  }  

  void render()
  {
    if ( !_rolled_over ) fill( PHYLO_TREE_COLOR );
    else fill( PHYLO_TREE_COLOR_ROLLED_OVER );

    noStroke();
    smooth();
    ellipse( _x, _y, 6, 6 );

    if ( _have_children )
    {
      // render the lines
      if ( !_rolled_over ) stroke( PHYLO_TREE_COLOR );
      else stroke( PHYLO_TREE_COLOR_ROLLED_OVER );
      strokeWeight( 1 );
      noSmooth();
      line( _x, _child[0].y(), _x, _child[1].y() );
      line( _x, _child[0].y(), _child[0].x(), _child[0].y() );
      line( _x, _child[1].y(), _child[1].x(), _child[1].y() );

      _child[0].render();
      _child[1].render();
    }
  }
}
























